import sys
import threading
import traceback
from collections import defaultdict
from datetime import datetime as Datetime

from dateutil.relativedelta import relativedelta

from flask import flash, g, make_response, render_template, request

import odio

from sqlalchemy import or_, select, true
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.computer import contract_func, forecast_date
from chellow.e.scenario import make_calcs, make_site_deltas
from chellow.models import (
    Bill,
    Contract,
    Era,
    MeasurementRequirement,
    RSession,
    Scenario,
    Site,
    SiteEra,
    Source,
    Ssc,
    Supply,
    Tpr,
    User,
)
from chellow.utils import (
    HH,
    PropDict,
    c_months_c,
    c_months_u,
    hh_format,
    hh_min,
    hh_range,
    make_val,
    req_bool,
    req_int,
    req_str,
    to_utc,
    utc_datetime_now,
)
from chellow.views import chellow_redirect


CATEGORY_ORDER = {None: 0, "unmetered": 1, "nhh": 2, "amr": 3, "hh": 4}
meter_order = {"hh": 0, "amr": 1, "nhh": 2, "unmetered": 3}


def write_spreadsheet(
    fl,
    compressed,
    site_rows,
    era_rows,
    read_rows,
):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, "1.2", compressed=compressed) as f:
        f.append_table("Site Level", site_rows)
        f.append_table("Era Level", era_rows)
        f.append_table("Normal Reads", read_rows)


def make_bill_row(titles, bill):
    return [bill.get(t) for t in titles]


def _add_bills(month_data, era, bills, chunk_start, chunk_finish):
    for bill in bills:
        contract = bill.batch.contract
        bill_role_code = contract.market_role.code
        bill_start = bill.start_date
        bill_finish = bill.finish_date
        bill_duration = (bill_finish - bill_start).total_seconds() + (30 * 60)
        overlap_duration = (
            min(bill_finish, chunk_finish) - max(bill_start, chunk_start)
        ).total_seconds() + (30 * 60)
        proportion = overlap_duration / bill_duration
        bill_prop_kwh = proportion * float(bill.kwh)
        bill_prop_net_gbp = proportion * float(bill.net)
        bill_prop_vat_gbp = proportion * float(bill.vat)
        bill_prop_gross_gbp = proportion * float(bill.gross)
        if bill_role_code == "X":
            role_name = "supplier"
            polarity = "export" if contract == era.exp_supplier_contract else "import"
        elif bill_role_code == "C":
            role_name = "dc"
            polarity = "import"
        elif bill_role_code == "M":
            role_name = "mop"
            polarity = "import"
        else:
            raise BadRequest("Role code not recognized.")

        month_data[f"billed-{polarity}-kwh"] += bill_prop_kwh
        month_data[f"billed-{polarity}-net-gbp"] += bill_prop_net_gbp
        month_data[f"billed-{polarity}-vat-gbp"] += bill_prop_vat_gbp
        month_data[f"billed-{polarity}-gross-gbp"] += bill_prop_gross_gbp
        month_data[f"billed-{role_name}-{polarity}-net-gbp"] += bill_prop_net_gbp
        month_data[f"billed-{role_name}-{polarity}-vat-gbp"] += bill_prop_vat_gbp
        month_data[f"billed-{role_name}-{polarity}-gross-gbp"] += bill_prop_gross_gbp


def _process_site(
    sess,
    report_context,
    forecast_from,
    start_date,
    finish_date,
    site,
    scenario_props,
    supply_ids,
    now,
    summary_titles,
    title_dict,
    era_rows,
    site_rows,
    data_source_bill,
):
    scenario_hh = scenario_props.get("hh_data", {})
    era_maps = scenario_props.get("era_maps", {})

    site_deltas = make_site_deltas(
        sess, report_context, site, scenario_hh, forecast_from, supply_ids
    )
    calcs, site_gen_types = make_calcs(
        sess,
        site,
        start_date,
        finish_date,
        supply_ids,
        site_deltas,
        forecast_from,
        report_context,
        era_maps,
        data_source_bill,
    )

    site_category = None
    site_sources = set()
    normal_reads = set()
    site_month_data = defaultdict(int)
    for i, (order, imp_mpan_core, exp_mpan_core, imp_ss, exp_ss) in enumerate(
        sorted(calcs, key=str)
    ):
        if imp_mpan_core == "displaced":
            month_data = {}
            for sname in (
                "import-grid",
                "export-grid",
                "import-gen",
                "export-gen",
                "import-3rd-party",
                "export-3rd-party",
                "msp",
                "used",
                "used-3rd-party",
                "billed-export",
            ):
                for xname in ("kwh", "net-gbp"):
                    month_data[f"{sname}-{xname}"] = 0
            month_data["billed-import-kwh"] = 0
            for suf in ("net-gbp", "vat-gbp", "gross-gbp"):
                month_data[f"billed-import-{suf}"] = 0
                month_data[f"billed-supplier-import-{suf}"] = None
                month_data[f"billed-dc-import-{suf}"] = None
                month_data[f"billed-mop-import-{suf}"] = None
                month_data[f"billed-export-{suf}"] = 0

            month_data["used-kwh"] = month_data["displaced-kwh"] = sum(
                hh["msp-kwh"] for hh in imp_ss.hh_data
            )

            disp_supplier_bill = imp_ss.supplier_bill

            gbp = disp_supplier_bill.get("net-gbp", 0)

            month_data["used-net-gbp"] = month_data["displaced-net-gbp"] = gbp

            out = (
                [
                    now,
                    None,
                    imp_ss.supplier_contract.name,
                    None,
                    None,
                    None,
                    imp_ss.era.meter_category,
                    "displaced",
                    None,
                    None,
                    None,
                    imp_ss.pc_code,
                    imp_ss.energisation_status_code,
                    imp_ss.gsp_group_code,
                    imp_ss.dno_code,
                    imp_ss.voltage_level_code,
                    imp_ss.is_substation,
                    imp_ss.llfc_code,
                    imp_ss.llfc.description,
                    None,
                    None,
                    None,
                    None,
                    site.code,
                    site.name,
                    "",
                    finish_date,
                ]
                + [month_data[t] for t in summary_titles]
                + [None]
                + [None] * len(title_dict["mop"])
                + [None]
                + [None] * len(title_dict["dc"])
                + [None]
                + make_bill_row(title_dict["imp-supplier"], disp_supplier_bill)
            )

            era_rows.append([make_val(v) for v in out])
            for k, v in month_data.items():
                if v is not None:
                    site_month_data[k] += v
        else:
            if imp_ss is None:
                source_code = exp_ss.source_code
                supply = exp_ss.supply
            else:
                source_code = imp_ss.source_code
                supply = imp_ss.supply

            site_sources.add(source_code)
            month_data = {}
            for name in (
                "import-grid",
                "export-grid",
                "import-gen",
                "export-gen",
                "import-3rd-party",
                "export-3rd-party",
                "displaced",
                "used",
                "used-3rd-party",
                "billed-export",
            ):
                for sname in ("kwh", "net-gbp"):
                    month_data[f"{name}-{sname}"] = 0
            for polarity in ("import", "export"):
                month_data[f"billed-{polarity}-kwh"] = 0
                for suf in ("net-gbp", "vat-gbp", "gross-gbp"):
                    month_data[f"billed-{polarity}-{suf}"] = 0
                    month_data[f"billed-supplier-{polarity}-{suf}"] = 0
                    month_data[f"billed-dc-{polarity}-{suf}"] = 0
                    month_data[f"billed-mop-{polarity}-{suf}"] = 0

            if imp_ss is not None:
                imp_supplier_contract = imp_ss.supplier_contract

                kwh = sum(hh["msp-kwh"] for hh in imp_ss.hh_data)
                imp_supplier_bill = imp_ss.supplier_bill

                gbp = imp_supplier_bill.get("net-gbp", 0)

                if source_code in ("grid", "gen-grid"):
                    month_data["import-grid-net-gbp"] += gbp
                    month_data["import-grid-kwh"] += kwh
                    month_data["used-net-gbp"] += gbp
                    month_data["used-kwh"] += kwh
                    if source_code == "gen-grid":
                        month_data["export-gen-kwh"] += kwh
                elif source_code == "3rd-party":
                    month_data["import-3rd-party-net-gbp"] += gbp
                    month_data["import-3rd-party-kwh"] += kwh
                    month_data["used-3rd-party-net-gbp"] += gbp
                    month_data["used-3rd-party-kwh"] += kwh
                    month_data["used-net-gbp"] += gbp
                    month_data["used-kwh"] += kwh
                elif source_code == "3rd-party-reverse":
                    month_data["export-3rd-party-net-gbp"] += gbp
                    month_data["export-3rd-party-kwh"] += kwh
                    month_data["used-3rd-party-net-gbp"] -= gbp
                    month_data["used-3rd-party-kwh"] -= kwh
                    month_data["used-net-gbp"] -= gbp
                    month_data["used-kwh"] -= kwh
                elif source_code == "gen":
                    month_data["import-gen-kwh"] += kwh

            if exp_ss is not None:
                exp_supplier_contract = exp_ss.supplier_contract

                kwh = sum(hh["msp-kwh"] for hh in exp_ss.hh_data)
                exp_supplier_bill = exp_ss.supplier_bill

                gbp = exp_supplier_bill.get("net-gbp", 0)

                if source_code in ("grid", "gen-grid"):
                    month_data["export-grid-net-gbp"] += gbp
                    month_data["export-grid-kwh"] += kwh
                    if source_code == "gen-grid":
                        month_data["import-gen-kwh"] += kwh

                elif source_code == "3rd-party":
                    month_data["export-3rd-party-net-gbp"] += gbp
                    month_data["export-3rd-party-kwh"] += kwh
                    month_data["used-3rd-party-net-gbp"] -= gbp
                    month_data["used-3rd-party-kwh"] -= kwh
                    month_data["used-net-gbp"] -= gbp
                    month_data["used-kwh"] -= kwh
                elif source_code == "3rd-party-reverse":
                    month_data["import-3rd-party-net-gbp"] += gbp
                    month_data["import-3rd-party-kwh"] += kwh
                    month_data["used-3rd-party-net-gbp"] += gbp
                    month_data["used-3rd-party-kwh"] += kwh
                    month_data["used-net-gbp"] += gbp
                    month_data["used-kwh"] += kwh
                elif source_code == "gen":
                    month_data["export-gen-kwh"] += kwh

            sss = exp_ss if imp_ss is None else imp_ss
            dc_contract = sss.dc_contract
            dc_vb_func = sss.contract_func(dc_contract, "virtual_bill")
            try:
                dc_vb_func(sss)
            except AttributeError as e:
                raise BadRequest(
                    f"Problem with virtual bill of DC contract {dc_contract.id} {e} "
                    f"{traceback.format_exc()}"
                )
            dc_bill = sss.dc_bill
            gbp = dc_bill["net-gbp"]

            mop_contract = sss.mop_contract
            mop_bill_function = sss.contract_func(mop_contract, "virtual_bill")
            try:
                mop_bill_function(sss)
            except (AttributeError, NameError) as e:
                raise BadRequest(
                    f"Problem with virtual bill of MOP contract {mop_contract.id} {e} "
                    f"{traceback.format_exc()}"
                )
            mop_bill = sss.mop_bill
            gbp += mop_bill["net-gbp"]

            if source_code in ("3rd-party", "3rd-party-reverse"):
                month_data["import-3rd-party-net-gbp"] += gbp
                month_data["used-3rd-party-net-gbp"] += gbp
            else:
                month_data["import-grid-net-gbp"] += gbp
            month_data["used-net-gbp"] += gbp

            generator_type = sss.generator_type_code
            if source_code in ("gen", "gen-grid"):
                site_gen_types.add(generator_type)

            era_category = sss.measurement_type
            if CATEGORY_ORDER[site_category] < CATEGORY_ORDER[era_category]:
                site_category = era_category

            era_associates = {
                s.site.code for s in sss.era.site_eras if not s.is_physical
            }
            bills = sess.scalars(
                select(Bill).where(
                    Bill.supply == supply,
                    Bill.start_date <= sss.finish_date,
                    Bill.finish_date >= sss.start_date,
                )
            ).all()
            _add_bills(month_data, sss.era, bills, sss.start_date, sss.finish_date)

            if imp_ss is None:
                imp_supplier_contract_name = imp_voltage_level_code = None
                imp_is_substation = imp_llfc_code = imp_llfc_description = None
                pc_code = exp_ss.pc_code
            else:
                if imp_supplier_contract is None:
                    imp_supplier_contract_name = ""
                else:
                    imp_supplier_contract_name = imp_supplier_contract.name
                pc_code = imp_ss.pc_code
                imp_voltage_level_code = imp_ss.voltage_level_code
                imp_is_substation = imp_ss.is_substation
                imp_llfc_code = imp_ss.llfc_code
                imp_llfc_description = imp_ss.llfc.description

            if exp_ss is None:
                exp_supplier_contract_name = exp_voltage_level_code = None
                exp_is_substation = exp_llfc_code = exp_llfc_description = None
            else:
                if exp_supplier_contract is None:
                    exp_supplier_contract_name = ""
                else:
                    exp_supplier_contract_name = exp_supplier_contract.name
                exp_voltage_level_code = exp_ss.voltage_level_code
                exp_is_substation = exp_ss.is_substation
                exp_llfc_code = exp_ss.llfc_code
                exp_llfc_description = exp_ss.llfc.description

            out = (
                [
                    now,
                    imp_mpan_core,
                    imp_supplier_contract_name,
                    exp_mpan_core,
                    exp_supplier_contract_name,
                    sss.era.start_date,
                    era_category,
                    source_code,
                    generator_type,
                    sss.supply_name,
                    sss.msn,
                    pc_code,
                    sss.energisation_status_code,
                    sss.gsp_group_code,
                    sss.dno_code,
                    imp_voltage_level_code,
                    imp_is_substation,
                    imp_llfc_code,
                    imp_llfc_description,
                    exp_voltage_level_code,
                    exp_is_substation,
                    exp_llfc_code,
                    exp_llfc_description,
                    site.code,
                    site.name,
                    ",".join(sorted(list(era_associates))),
                    finish_date,
                ]
                + [month_data[t] for t in summary_titles]
                + [None]
                + make_bill_row(title_dict["mop"], mop_bill)
                + [None]
                + make_bill_row(title_dict["dc"], dc_bill)
            )
            if imp_ss is None:
                out += [None] * (len(title_dict["imp-supplier"]) + 1)
            else:
                out += [None] + make_bill_row(
                    title_dict["imp-supplier"], imp_supplier_bill
                )
                for n in imp_ss.normal_reads:
                    normal_reads.add((imp_mpan_core, n))
            if exp_ss is not None:
                out += [None] + make_bill_row(
                    title_dict["exp-supplier"], exp_supplier_bill
                )

            for k, v in month_data.items():
                site_month_data[k] += v
            era_rows.append([make_val(v) for v in out])

    q = select(Supply).join(Era).join(SiteEra).where(SiteEra.site == site).distinct()
    if supply_ids is not None:
        q = q.where(Supply.id.in_(supply_ids))
    for supply in sess.execute(q).scalars():
        last_era = (
            sess.execute(
                select(Era).where(Era.supply == supply).order_by(Era.start_date.desc())
            )
            .scalars()
            .first()
        )
        if last_era.finish_date is not None and last_era.start_date <= finish_date:
            site_era = sess.execute(
                select(SiteEra).where(
                    SiteEra.era == last_era,
                    SiteEra.is_physical == true(),
                    SiteEra.site == site,
                )
            ).scalar_one_or_none()
            if site_era is not None:
                chunk_start = max(start_date, last_era.finish_date + HH)
                chunk_finish = finish_date
                bills = (
                    sess.execute(
                        select(Bill).where(
                            Bill.supply == supply,
                            Bill.start_date <= chunk_finish,
                            Bill.finish_date >= chunk_start,
                        )
                    )
                    .scalars()
                    .all()
                )
                if len(bills) > 0:
                    month_data = {}
                    for name in (
                        "import-grid",
                        "export-grid",
                        "import-gen",
                        "export-gen",
                        "import-3rd-party",
                        "export-3rd-party",
                        "displaced",
                        "used",
                        "used-3rd-party",
                        "billed-export",
                    ):
                        for sname in ("kwh", "net-gbp"):
                            month_data[f"{name}-{sname}"] = 0
                    month_data["billed-import-kwh"] = 0
                    for suf in ("net-gbp", "vat-gbp", "gross-gbp"):
                        month_data[f"billed-import-{suf}"] = 0
                        month_data[f"billed-supplier-import-{suf}"] = 0
                        month_data[f"billed-dc-import-{suf}"] = 0
                        month_data[f"billed-mop-import-{suf}"] = 0

                    _add_bills(month_data, last_era, bills, chunk_start, chunk_finish)

                    imp_supplier_contract = last_era.imp_supplier_contract
                    imp_llfc = last_era.imp_llfc
                    if imp_llfc is None:
                        imp_llfc_code = imp_voltage_level_code = None
                        imp_is_substation = imp_llfc_description = None
                    else:
                        imp_llfc_code = imp_llfc.code
                        imp_voltage_level_code = imp_llfc.voltage_level.code
                        imp_is_substation = imp_llfc.is_substation
                        imp_llfc_description = imp_llfc.description

                    exp_supplier_contract = last_era.exp_supplier_contract
                    exp_llfc = last_era.exp_llfc
                    if exp_llfc is None:
                        exp_llfc_code = exp_voltage_level_code = None
                        exp_is_substation = exp_llfc_description = None
                    else:
                        exp_llfc_code = exp_llfc.code
                        exp_voltage_level_code = exp_llfc.voltage_level.code
                        exp_is_substation = exp_llfc.is_substation
                        exp_llfc_description = exp_llfc.description
                    out = [
                        now,
                        last_era.imp_mpan_core,
                        (
                            None
                            if imp_supplier_contract is None
                            else imp_supplier_contract.name
                        ),
                        last_era.exp_mpan_core,
                        (
                            None
                            if exp_supplier_contract is None
                            else exp_supplier_contract.name
                        ),
                        chunk_start,
                        last_era.meter_category,
                        last_era.supply.source.code,
                        None,
                        last_era.supply.name,
                        last_era.msn,
                        last_era.pc.code,
                        last_era.energisation_status.code,
                        last_era.supply.gsp_group.code,
                        last_era.supply.dno.dno_code,
                        imp_voltage_level_code,
                        imp_is_substation,
                        imp_llfc_code,
                        imp_llfc_description,
                        exp_voltage_level_code,
                        exp_is_substation,
                        exp_llfc_code,
                        exp_llfc_description,
                        site.code,
                        site.name,
                        None,
                        finish_date,
                    ] + [month_data[t] for t in summary_titles]

                    era_rows.append([make_val(v) for v in out])
                    for k, v in month_data.items():
                        if v is not None:
                            site_month_data[k] += v
        first_era = (
            sess.execute(
                select(Era).where(Era.supply == supply).order_by(Era.start_date)
            )
            .scalars()
            .first()
        )
        if first_era.start_date > start_date:
            site_era = sess.execute(
                select(SiteEra).where(
                    SiteEra.era == first_era,
                    SiteEra.is_physical == true(),
                    SiteEra.site == site,
                )
            ).scalar_one_or_none()
            if site_era is not None:
                chunk_start = start_date
                chunk_finish = hh_min(finish_date, first_era.start_date - HH)
                bills = (
                    sess.execute(
                        select(Bill).where(
                            Bill.supply == supply,
                            Bill.start_date <= chunk_finish,
                            Bill.finish_date >= chunk_start,
                        )
                    )
                    .scalars()
                    .all()
                )
                if len(bills) > 0:
                    month_data = {}
                    for name in (
                        "import-grid",
                        "export-grid",
                        "import-gen",
                        "export-gen",
                        "import-3rd-party",
                        "export-3rd-party",
                        "displaced",
                        "used",
                        "used-3rd-party",
                        "billed-export",
                    ):
                        for sname in ("kwh", "net-gbp"):
                            month_data[f"{name}-{sname}"] = 0
                    month_data["billed-import-kwh"] = 0
                    for suf in ("net-gbp", "vat-gbp", "gross-gbp"):
                        month_data[f"billed-import-{suf}"] = 0
                        month_data[f"billed-supplier-import-{suf}"] = 0
                        month_data[f"billed-dc-import-{suf}"] = 0
                        month_data[f"billed-mop-import-{suf}"] = 0
                    _add_bills(month_data, first_era, bills, chunk_start, chunk_finish)

                    imp_supplier_contract = first_era.imp_supplier_contract
                    imp_llfc = last_era.imp_llfc
                    if imp_llfc is None:
                        imp_llfc_code = imp_voltage_level_code = None
                        imp_is_substation = imp_llfc_description = None
                    else:
                        imp_llfc_code = imp_llfc.code
                        imp_voltage_level_code = imp_llfc.voltage_level.code
                        imp_is_substation = imp_llfc.is_substation
                        imp_llfc_description = imp_llfc.description

                    exp_llfc = last_era.exp_llfc
                    if exp_llfc is None:
                        exp_llfc_code = exp_voltage_level_code = None
                        exp_is_substation = exp_llfc_description = None
                    else:
                        exp_llfc_code = exp_llfc.code
                        exp_voltage_level_code = exp_llfc.voltage_level.code
                        exp_is_substation = exp_llfc.is_substation
                        exp_llfc_description = exp_llfc.description
                    exp_supplier_contract = first_era.exp_supplier_contract
                    out = [
                        now,
                        last_era.imp_mpan_core,
                        (
                            None
                            if imp_supplier_contract is None
                            else imp_supplier_contract.name
                        ),
                        last_era.exp_mpan_core,
                        (
                            None
                            if exp_supplier_contract is None
                            else exp_supplier_contract.name
                        ),
                        None,
                        last_era.meter_category,
                        last_era.supply.source.code,
                        None,
                        last_era.supply.name,
                        last_era.msn,
                        last_era.pc.code,
                        last_era.energisation_status.code,
                        last_era.supply.gsp_group.code,
                        last_era.supply.dno.dno_code,
                        imp_voltage_level_code,
                        imp_is_substation,
                        imp_llfc_code,
                        imp_llfc_description,
                        exp_voltage_level_code,
                        exp_is_substation,
                        exp_llfc_code,
                        exp_llfc_description,
                        site.code,
                        site.name,
                        None,
                        finish_date,
                    ] + [month_data[t] for t in summary_titles]

                    era_rows.append([make_val(v) for v in out])
                    for k, v in month_data.items():
                        if v is not None:
                            site_month_data[k] += v
    site_row = [
        now,
        site.code,
        site.name,
        ", ".join(
            s.code for s in site.find_linked_sites(sess, start_date, finish_date)
        ),
        finish_date,
        site_category,
        ", ".join(sorted(list(site_sources))),
        ", ".join(sorted(list(site_gen_types))),
    ] + [site_month_data[k] for k in summary_titles]

    site_rows.append([make_val(v) for v in site_row])
    return normal_reads


class Object:
    pass


def content(scenario_props, base_name, user_id, compression, now):
    report_context = {}

    try:
        comp = report_context["computer"]
    except KeyError:
        comp = report_context["computer"] = {}

    try:
        rate_cache = comp["rates"]
    except KeyError:
        rate_cache = comp["rates"] = {}

    try:
        ind_cont = report_context["contract_names"]
    except KeyError:
        ind_cont = report_context["contract_names"] = {}

    sess = rf = None
    site_rows = []
    era_rows = []
    normal_read_rows = []
    try:
        with RSession() as sess:
            start_year = scenario_props["scenario_start_year"]
            start_month = scenario_props["scenario_start_month"]
            months = scenario_props["scenario_duration"]

            month_pairs = list(
                c_months_u(
                    start_year=start_year, start_month=start_month, months=months
                )
            )
            start_date_utc = month_pairs[0][0]
            finish_date_utc = month_pairs[-1][-1]

            base_name.append(
                hh_format(start_date_utc)
                .replace(" ", "_")
                .replace(":", "")
                .replace("-", "")
            )

            base_name.append("for")
            base_name.append(str(months))
            base_name.append("months")

            if "forecast_from" in scenario_props:
                forecast_from = scenario_props["forecast_from"]
            else:
                forecast_from = None

            if forecast_from is None:
                forecast_from = forecast_date()
            else:
                forecast_from = to_utc(forecast_from)

            sites = sess.query(Site).distinct().order_by(Site.code)

            mpan_cores = scenario_props.get("mpan_cores")
            supply_ids = None
            if mpan_cores is not None:
                supply_ids = []
                for mpan_core in mpan_cores:
                    supply = Supply.get_by_mpan_core(sess, mpan_core)
                    supply_ids.append(supply.id)

                if len(supply_ids) == 1:
                    base_name.append("supply")
                    base_name.append(str(supply.id))
                else:
                    base_name.append("mpan_cores")

                sites = (
                    sites.join(SiteEra)
                    .join(Era)
                    .join(Supply)
                    .where(Supply.id.in_(supply_ids))
                )

            site_codes = scenario_props.get("site_codes")
            if site_codes is not None:
                if len(site_codes) == 1:
                    base_name.append("site")
                    base_name.append(site_codes[0])
                else:
                    base_name.append("sitecodes")
                sites = sites.where(Site.code.in_(site_codes))

            user = User.get_by_id(sess, user_id)

            rf = open_file("_".join(base_name) + ".ods", user, mode="wb")

            for rate_script in scenario_props.get("rates", []):
                contract_id = rate_script["contract_id"]
                try:
                    cont_cache = rate_cache[contract_id]
                except KeyError:
                    cont_cache = rate_cache[contract_id] = {}

                try:
                    rate_script_start = rate_script["start_date"]
                except KeyError:
                    raise BadRequest(
                        f"Problem in the scenario properties. Can't find the "
                        f"'start_date' key of the contract {contract_id} in the "
                        f"'rates' map."
                    )

                try:
                    rate_script_start = rate_script["start_date"]
                except KeyError:
                    raise BadRequest(
                        f"Problem in the scenario properties. Can't find the "
                        f"'start_date' key of the contract {contract_id} in the "
                        f"'rates' map."
                    )

                props = PropDict("scenario properties", rate_script["script"])
                for dt in hh_range(
                    report_context, rate_script_start, rate_script["finish_date"]
                ):
                    cont_cache[dt] = props

            for rate_script in scenario_props.get("industry_rates", []):
                contract_name = rate_script["contract_name"]
                try:
                    cont_cache = ind_cont[contract_name]
                except KeyError:
                    cont_cache = ind_cont[contract_name] = {}

                rfinish = rate_script["finish_date"]
                if rfinish is None:
                    raise BadRequest(
                        f"For the industry rate {contract_name} the finish_date can't "
                        f"be null."
                    )
                for dt in hh_range(report_context, rate_script["start_date"], rfinish):
                    cont_cache[dt] = PropDict(
                        "scenario properties", rate_script["script"]
                    )

            by_hh = scenario_props.get("by_hh", False)

            era_header_titles = [
                "creation-date",
                "imp-mpan-core",
                "imp-supplier-contract",
                "exp-mpan-core",
                "exp-supplier-contract",
                "era-start-date",
                "metering-type",
                "source",
                "generator-type",
                "supply-name",
                "msn",
                "pc",
                "energisation-status",
                "gsp-group",
                "dno",
                "imp-voltage-level",
                "imp-is-substation",
                "imp-llfc-code",
                "imp-llfc-description",
                "exp-voltage-level",
                "exp-is-substation",
                "exp-llfc-code",
                "exp-llfc-description",
                "site-id",
                "site-name",
                "associated-site-ids",
                "month",
            ]
            site_header_titles = [
                "creation-date",
                "site-id",
                "site-name",
                "associated-site-ids",
                "month",
                "metering-type",
                "sources",
                "generator-types",
            ]
            summary_titles = [
                "import-grid-kwh",
                "export-grid-kwh",
                "import-gen-kwh",
                "export-gen-kwh",
                "import-3rd-party-kwh",
                "export-3rd-party-kwh",
                "displaced-kwh",
                "used-kwh",
                "used-3rd-party-kwh",
                "import-grid-net-gbp",
                "export-grid-net-gbp",
                "import-gen-net-gbp",
                "export-gen-net-gbp",
                "import-3rd-party-net-gbp",
                "export-3rd-party-net-gbp",
                "displaced-net-gbp",
                "used-net-gbp",
                "used-3rd-party-net-gbp",
                "billed-import-kwh",
                "billed-import-net-gbp",
                "billed-import-vat-gbp",
                "billed-import-gross-gbp",
                "billed-supplier-import-net-gbp",
                "billed-supplier-import-vat-gbp",
                "billed-supplier-import-gross-gbp",
                "billed-dc-import-net-gbp",
                "billed-dc-import-vat-gbp",
                "billed-dc-import-gross-gbp",
                "billed-mop-import-net-gbp",
                "billed-mop-import-vat-gbp",
                "billed-mop-import-gross-gbp",
                "billed-export-kwh",
                "billed-export-net-gbp",
            ]

            title_dict = {}
            for cont_type, con_attr in (
                ("mop", Era.mop_contract),
                ("dc", Era.dc_contract),
                ("imp-supplier", Era.imp_supplier_contract),
                ("exp-supplier", Era.exp_supplier_contract),
            ):
                titles = []
                title_dict[cont_type] = titles
                conts = (
                    sess.query(Contract)
                    .join(con_attr)
                    .join(Era.supply)
                    .join(Source)
                    .filter(
                        Era.start_date <= finish_date_utc,
                        or_(
                            Era.finish_date == null(), Era.finish_date >= start_date_utc
                        ),
                    )
                    .distinct()
                    .order_by(Contract.id)
                )
                if supply_ids is not None:
                    conts = conts.where(Supply.id.in_(supply_ids))
                for cont in conts:
                    title_func = contract_func(
                        report_context, cont, "virtual_bill_titles"
                    )
                    if title_func is None:
                        raise Exception(
                            f"For the contract {cont.name} there doesn't seem to be a "
                            f"'virtual_bill_titles' function."
                        )
                    for title in title_func():
                        if title not in titles:
                            titles.append(title)

            tpr_query = (
                sess.query(Tpr)
                .join(MeasurementRequirement)
                .join(Ssc)
                .join(Era)
                .filter(
                    Era.start_date <= finish_date_utc,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date_utc),
                )
                .order_by(Tpr.code)
                .distinct()
            )
            for tpr in tpr_query.filter(Era.imp_supplier_contract != null()):
                for suffix in ("-kwh", "-rate", "-gbp"):
                    title_dict["imp-supplier"].append(tpr.code + suffix)
            for tpr in tpr_query.filter(Era.exp_supplier_contract != null()):
                for suffix in ("-kwh", "-rate", "-gbp"):
                    title_dict["exp-supplier"].append(tpr.code + suffix)

            era_titles = (
                era_header_titles
                + summary_titles
                + [None]
                + ["mop-" + t for t in title_dict["mop"]]
                + [None]
                + ["dc-" + t for t in title_dict["dc"]]
                + [None]
                + ["imp-supplier-" + t for t in title_dict["imp-supplier"]]
                + [None]
                + ["exp-supplier-" + t for t in title_dict["exp-supplier"]]
            )
            site_rows.append(site_header_titles + summary_titles)
            era_rows.append(era_titles)

            sites = sites.all()
            normal_reads = set()

            for month_start, month_finish in month_pairs:
                if scenario_props.get("is_bill_check", False):
                    data_source_bill = Object()
                    data_source_bill.start_date = month_start
                    data_source_bill.finish_date = month_finish
                else:
                    data_source_bill = None
                for site in sites:
                    if by_hh:
                        sf = [
                            (d, d)
                            for d in hh_range(report_context, month_start, month_finish)
                        ]
                    else:
                        sf = [(month_start, month_finish)]

                    for start, finish in sf:
                        try:
                            normal_reads = normal_reads | _process_site(
                                sess,
                                report_context,
                                forecast_from,
                                start,
                                finish,
                                site,
                                scenario_props,
                                supply_ids,
                                now,
                                summary_titles,
                                title_dict,
                                era_rows,
                                site_rows,
                                data_source_bill,
                            )
                        except BadRequest as e:
                            raise BadRequest(f"Site Code {site.code}: {e.description}")

                    sess.rollback()  # Evict from cache

                normal_read_rows = [["mpan_core", "date", "msn", "type", "registers"]]
                for mpan_core, r in sorted(list(normal_reads)):
                    row = [mpan_core, r.date, r.msn, r.type] + list(r.reads)
                    normal_read_rows.append(row)

                write_spreadsheet(
                    rf, compression, site_rows, era_rows, normal_read_rows
                )
    except BadRequest as e:
        msg = e.description + traceback.format_exc()
        sys.stderr.write(msg + "\n")
        site_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, site_rows, era_rows, normal_read_rows)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        site_rows.append(["Problem " + msg])
        if rf is None:
            msg = traceback.format_exc()
            ef = open_file("error.txt", None, mode="w")
            ef.write(msg + "\n")
            ef.close()
        else:
            write_spreadsheet(rf, compression, site_rows, era_rows, normal_read_rows)
    finally:
        if rf is not None:
            rf.close()


def do_post(sess):
    base_name = []
    now = utc_datetime_now()

    if "scenario_id" in request.values:
        scenario_id = req_int("scenario_id")
        scenario = Scenario.get_by_id(sess, scenario_id)
        scenario_props = scenario.props
        base_name.append(scenario.name)
    else:
        scenario_props = {}
        base_name.append("monthly_duration")

    if "finish_year" in request.values:
        year = req_int("finish_year")
        month = req_int("finish_month")
        months = req_int("months")
        start_date, _ = next(
            c_months_c(finish_year=year, finish_month=month, months=months)
        )
        scenario_props["scenario_start_year"] = start_date.year
        scenario_props["scenario_start_month"] = start_date.month
        scenario_props["scenario_duration"] = months

    scenario_props["by_hh"] = req_bool("by_hh")

    try:
        if "site_id" in request.values:
            site_id = req_int("site_id")
            scenario_props["site_codes"] = [Site.get_by_id(sess, site_id).code]

        if "site_codes" in request.values:
            site_codes_raw_str = req_str("site_codes")
            site_codes_str = site_codes_raw_str.strip()
            if len(site_codes_str) > 0:
                site_codes = []

                for site_code in site_codes_str.splitlines():
                    Site.get_by_code(sess, site_code)  # Check valid
                    site_codes.append(site_code)

                scenario_props["site_codes"] = site_codes

        if "supply_id" in request.values:
            supply_id = req_int("supply_id")
            supply = Supply.get_by_id(sess, supply_id)
            era = supply.eras[-1]
            imp_mpan_core, exp_mpan_core = era.imp_mpan_core, era.exp_mpan_core
            scenario_props["mpan_cores"] = [
                exp_mpan_core if imp_mpan_core is None else imp_mpan_core
            ]

        if "compression" in request.values:
            compression = req_bool("compression")
        else:
            compression = True

        user = g.user

        args = (scenario_props, base_name, user.id, compression, now)
        threading.Thread(target=content, args=args).start()
        return chellow_redirect("/downloads", 303)
    except BadRequest as e:
        flash(e.description)
        if "scenario_id" in request.values:
            return make_response(
                render_template("e/scenario.html", scenario=scenario), 400
            )
        else:
            now = Datetime.utcnow()
            month_start = Datetime(now.year, now.month, 1) - relativedelta(months=1)
            month_finish = Datetime(now.year, now.month, 1) - HH
            return make_response(
                render_template(
                    "e/ods_monthly_duration.html",
                    month_start=month_start,
                    month_finish=month_finish,
                ),
                400,
            )



import atexit
import collections
import threading
import traceback
from datetime import timedelta
from importlib import import_module

import requests

from werkzeug.exceptions import BadRequest

from chellow.models import (
    Contract,
    Session,
)
from chellow.utils import ct_datetime_now, hh_format, utc_datetime_now


importer = None


def api_records(log, s, resource_id, skip=0):
    url = f"https://dp.lowcarboncontracts.uk/datastore/dump/{resource_id}"
    params = {"format": "json"}
    res = s.get(url, params=params)
    log(f"Requested URL {res.url}")
    try:
        res_j = res.json()
    except requests.exceptions.JSONDecodeError as e:
        raise BadRequest(
            f"Couldn't parse as JSON the content from {url} with error {e}: "
            f"{res.text}"
        )

    if "success" in res_j and not res_j["success"]:
        raise BadRequest(res_j)

    field_titles = [f["id"] for f in res_j["fields"]]

    for record in res_j["records"][skip:]:
        yield {k: v for k, v in zip(field_titles, record)}


def run_import(sess, log, set_progress):
    log("Starting to import data from the Low Carbon Contracts Company")
    s = requests.Session()
    s.verify = False

    for mod_name in ("chellow.e.cfd",):
        mod = import_module(mod_name)
        mod.lcc_import(sess, log, set_progress, s)


LAST_RUN_KEY = "last_run"
GLOBAL_ALERT = (
    "There's a problem with a <a href='/e/lcc'>Low Carbon Contracts import</a>."
)
LCC_STATE_KEY = "lcc"
DELAY_DAYS = 7


class LowCarbonContracts(threading.Thread):
    def __init__(self):
        super().__init__(name="Low Carbon Contracts")
        self.messages = collections.deque(maxlen=500)
        self.progress = ""
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def log(self, message):
        self.messages.appendleft(
            f"{ct_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def set_progress(self, progress):
        self.progress = progress

    def run(self):
        while not self.stopped.is_set():
            with Session() as sess:
                try:
                    config = Contract.get_non_core_by_name(sess, "configuration")
                    state = config.make_state()
                    lcc_state = state.get(LCC_STATE_KEY, {})
                except BaseException as e:
                    msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                    self.log(f"{msg}{traceback.format_exc()}")
                    self.global_alert = GLOBAL_ALERT
                    sess.rollback()

            last_run = lcc_state.get(LAST_RUN_KEY)
            if last_run is None or utc_datetime_now() - last_run > timedelta(
                days=DELAY_DAYS
            ):
                self.going.set()

            if self.going.is_set():
                self.global_alert = None
                with Session() as sess:
                    try:
                        config = Contract.get_non_core_by_name(sess, "configuration")
                        state = config.make_state()
                        try:
                            lcc_state = state[LCC_STATE_KEY]
                        except KeyError:
                            lcc_state = state[LCC_STATE_KEY] = {}

                        lcc_state[LAST_RUN_KEY] = utc_datetime_now()
                        config.update_state(state)
                        sess.commit()
                        run_import(sess, self.log, self.set_progress)
                    except BaseException as e:
                        msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                        self.log(f"{msg}{traceback.format_exc()}")
                        self.global_alert = GLOBAL_ALERT
                        sess.rollback()
                    finally:
                        self.going.clear()
                        self.log("Finished importing Low Carbon Contracts data.")

            else:
                self.log(
                    f"The importer was last run at {hh_format(last_run)}. There will "
                    f"be another import when {DELAY_DAYS} days have elapsed since the "
                    f"last run."
                )
                self.going.wait(60 * 60)


def get_importer():
    return importer


def startup():
    global importer
    importer = LowCarbonContracts()
    importer.start()


@atexit.register
def shutdown():
    if importer is not None:
        importer.stop()



import csv
import decimal
import itertools
from datetime import timedelta

from werkzeug.exceptions import BadRequest

from chellow.utils import parse_mpan_core, utc_datetime


def create_parser(reader, mpan_map, messages):
    return HhParserBglobal(reader, mpan_map, messages)


class HhParserBglobal:
    def __init__(self, reader, mpan_map, messages):
        self.shredder = zip(itertools.count(1), csv.reader(reader))
        self.line_number, self.values = next(self.shredder)
        self.mpan_map = mpan_map
        self.col_idx = 0

    def __iter__(self):
        return self

    def __next__(self):
        datum = None
        try:
            while datum is None:
                if self.col_idx > 50:
                    self.line_number, self.values = next(self.shredder)
                    if len(self.values) == 0:
                        continue
                    self.col_idx = 0

                if self.col_idx == 0:
                    try:
                        self.core = self.values[self.col_idx]
                    except KeyError:
                        raise BadRequest(
                            "There doesn't seem to be an MPAN Core at the "
                            "beginning of this line. "
                        )
                    self.core = self.mpan_map.get(self.core, self.core)
                    self.core = parse_mpan_core(self.core)
                elif self.col_idx == 2:
                    day, month, year = map(int, self.values[self.col_idx].split("/"))
                    self.date = utc_datetime(year, month, day)
                elif 2 < self.col_idx < len(self.values):
                    hh_value = self.values[self.col_idx].strip()
                    mins = 30 * (self.col_idx - 3)
                    if len(hh_value) > 0:
                        datum = {
                            "mpan_core": self.core,
                            "channel_type": "ACTIVE",
                            "start_date": self.date + timedelta(minutes=mins),
                            "value": decimal.Decimal(hh_value),
                            "status": "A",
                        }

                self.col_idx += 1
        except BadRequest as e:
            e.description = "".join(
                "Problem at line number: ", str(self.line_number), ": ", e.description
            )
            raise e
        return datum

    def close(self):
        self.shredder.close()



from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.edi_lib import parse_edi, to_date, to_decimal
from chellow.utils import HH, to_ct, to_utc

# From Total 2024-05-29
# ReadType == "C"  = "00"
# ReadType == "N" = "00"
# ReadType == "A" = "00"
# ReadType == "PA" = "00"
# ReadType == "S"  = "00"
# ReadType == "I"  = "00"
# ReadType == "E" = "02"
# ReadType == "AE"  = "02"
# ReadType == "M"  = "02"
# ReadType == "PE"  = "02"

READ_TYPE_MAP = {"00": "A", "02": "E"}


SUPPLIER_CODE_MAP = {
    "STD": ("standing", Decimal("1000"), Decimal("1")),
    "MET": ("commodity", Decimal("100000"), Decimal("1000")),
    "CCL": ("ccl", Decimal("100000"), Decimal("1000")),
}

UNIT_MAP = {"M3": "M3", "HH": "HCUF", "HCUF": "HCUF"}


def _to_finish_date(date_str):
    if len(date_str) == 0:
        return None
    return to_utc(
        to_ct(Datetime.strptime(date_str, "%y%m%d") + relativedelta(days=1) - HH)
    )


def _process_ADJ(elements, headers):
    adjf = elements["ADJF"]
    if adjf[0] == "CV":
        headers["reads"][-1]["calorific_value"] = Decimal(adjf[1]) / Decimal(100000)


def _process_BCD(elements, headers):
    ivdt = elements["IVDT"]
    headers["issue_date"] = to_date(ivdt[0])

    invn = elements["INVN"]
    headers["reference"] = invn[0]

    btcd = elements["BTCD"]
    headers["bill_type_code"] = btcd[0]


def _process_MHD(elements, headers):
    headers.clear()

    typ = elements["TYPE"]
    message_type = headers["message_type"] = typ[0]
    if message_type == "UTLBIL":
        headers["reads"] = []
        headers["raw_lines"] = []
        headers["breakdown"] = defaultdict(int, {"units_consumed": Decimal(0)})
        headers["kwh"] = Decimal("0.00")
        headers["net"] = Decimal("0.00")
        headers["vat"] = Decimal("0.00")
        headers["gross"] = Decimal("0.00")


NUCT_LOOKUP = {"DAY": "days", "KWH": "kwh"}


def _process_CCD2(elements, headers):
    breakdown = headers["breakdown"]
    ccde = elements["CCDE"]
    nuct = elements["NUCT"]
    mtnr = elements["MTNR"]
    conb = elements["CONB"]
    adjf = elements["ADJF"]
    prdt = elements["PRDT"]
    pvdt = elements["PVDT"]
    prrd = elements["PRRD"]
    cppu = elements["CPPU"]
    ctot = elements["CTOT"]
    csdt = elements["CSDT"]
    cedt = elements["CEDT"]
    mloc = elements["MLOC"]

    ccde_supplier_code = ccde[2]
    tpref, rate_divisor, units_divisor = SUPPLIER_CODE_MAP[ccde_supplier_code]

    rate_str = cppu[0]

    rate_key = f"{tpref}_rate"
    if rate_key not in breakdown:
        breakdown[rate_key] = set()
    rate = Decimal(rate_str) / rate_divisor
    breakdown[rate_key].add(rate)

    breakdown[f"{tpref}_gbp"] += to_decimal(ctot) / Decimal("100")

    suff = NUCT_LOOKUP[nuct[1].strip()]
    key = f"{tpref}_{suff}"
    units_billed = to_decimal(nuct) / units_divisor
    breakdown[key] += units_billed

    if tpref == "standing" or "start_date" not in headers:
        headers["start_date"] = to_date(csdt[0])

    if tpref == "standing" or "finish_date" not in headers:
        headers["finish_date"] = _to_finish_date(cedt[0])

    if "mprn" not in headers:
        headers["mprn"] = mloc[0]

    if len(conb) > 0 and len(conb[0]) > 0:
        headers["breakdown"]["units_consumed"] += to_decimal(conb) / Decimal("1000")

    if tpref == "commodity":
        headers["kwh"] += units_billed

    if len(prrd) > 0 and len(prrd[0]) > 0:
        pres_read_date = to_date(prdt[0])
        prev_read_date = to_date(pvdt[0])

        pres_read_value = Decimal(prrd[0])
        pres_read_type = READ_TYPE_MAP[prrd[1]]
        prev_read_value = Decimal(prrd[2])
        prev_read_type = READ_TYPE_MAP[prrd[3]]
        msn = mtnr[0]
        unit = UNIT_MAP[conb[1]]
        correction_factor = Decimal(adjf[1]) / Decimal(100000)

        headers["reads"].append(
            {
                "msn": msn,
                "unit": unit,
                "correction_factor": correction_factor,
                "prev_date": prev_read_date,
                "prev_value": prev_read_value,
                "prev_type_code": prev_read_type,
                "pres_date": pres_read_date,
                "pres_value": pres_read_value,
                "pres_type_code": pres_read_type,
            }
        )


def _process_MTR(elements, headers):
    if headers["message_type"] == "UTLBIL":
        breakdown = headers["breakdown"]
        for k, v in tuple(breakdown.items()):
            if isinstance(v, set):
                breakdown[k] = sorted(v)

        return {
            "raw_lines": "\n".join(headers["raw_lines"]),
            "mprn": headers["mprn"],
            "reference": headers["reference"],
            "account": headers["mprn"],
            "reads": headers["reads"],
            "kwh": headers["kwh"],
            "breakdown": headers["breakdown"],
            "net_gbp": headers["net"],
            "vat_gbp": headers["vat"],
            "gross_gbp": headers["gross"],
            "bill_type_code": headers["bill_type_code"],
            "start_date": headers["start_date"],
            "finish_date": headers["finish_date"],
            "issue_date": headers["issue_date"],
        }


def _process_VAT(elements, headers):
    breakdown = headers["breakdown"]
    vatp = elements["VATP"]
    if "vat" in breakdown:
        vat = breakdown["vat"]
    else:
        vat = breakdown["vat"] = {}

    vat_perc = to_decimal(vatp) / Decimal(1000)
    try:
        vat_bd = vat[vat_perc]
    except KeyError:
        vat_bd = vat[vat_perc] = defaultdict(int)

    uvtt = elements["UVTT"]
    vat_gbp = to_decimal(uvtt) / Decimal("100")

    vat_bd["vat"] += vat_gbp

    uvla = elements["UVLA"]
    net_gbp = to_decimal(uvla) / Decimal("100")
    vat_bd["net"] += net_gbp
    headers["net"] += net_gbp
    headers["vat"] += vat_gbp
    ucsi = elements["UCSI"]
    headers["gross"] += to_decimal(ucsi) / Decimal("100")


def _process_NOOP(elements, headers):
    pass


CODE_FUNCS = {
    "ADJ": _process_ADJ,
    "BCD": _process_BCD,
    "BTL": _process_NOOP,
    "CCD1": _process_NOOP,
    "CCD2": _process_CCD2,
    "CCD3": _process_NOOP,
    "CCD4": _process_NOOP,
    "CDT": _process_NOOP,
    "CLO": _process_NOOP,
    "END": _process_NOOP,
    "FIL": _process_NOOP,
    "MHD": _process_MHD,
    "MTR": _process_MTR,
    "SDT": _process_NOOP,
    "STX": _process_NOOP,
    "TYP": _process_NOOP,
    "TTL": _process_NOOP,
    "VAT": _process_VAT,
    "VTS": _process_NOOP,
}


class Parser:
    def __init__(self, file_bytes):
        self.edi_str = str(file_bytes, "utf-8", errors="ignore")
        self.line_number = None

    def make_raw_bills(self):
        bills = []
        headers = {}
        bill = None
        for self.line_number, line, seg_name, elements in parse_edi(self.edi_str):
            try:
                func = CODE_FUNCS[seg_name]
            except KeyError:
                raise BadRequest(f"Code {seg_name} not recognized.")
            try:
                bill = func(elements, headers)
            except BaseException as e:
                raise Exception(f"Propblem with segment {line}: {e}") from e

            if "raw_lines" in headers:
                headers["raw_lines"].append(line)
            if bill is not None:
                bills.append(bill)

        return bills



import csv
import sys
import threading
import traceback

from dateutil.relativedelta import relativedelta

from flask import g, request

from sqlalchemy import or_, text
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import func, null

from werkzeug.exceptions import BadRequest

import chellow.e.duos
import chellow.e.tnuos
from chellow.dloads import open_file
from chellow.e.computer import SupplySource
from chellow.models import (
    Batch,
    Bill,
    Era,
    GeneratorType,
    MeasurementRequirement,
    MtcParticipant,
    RSession,
    ReadType,
    RegisterRead,
    SiteEra,
    Supply,
    User,
)
from chellow.utils import (
    CHANNEL_TYPES,
    HH,
    csv_make_val,
    hh_format,
    hh_min,
    parse_mpan_core,
    req_date,
    req_int,
    req_str,
)
from chellow.views import chellow_redirect


def content(user_id, date, supply_id, mpan_cores):
    try:
        with RSession() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("supplies_snapshot.csv", user, mode="w", newline="")
            _process(sess, f, date, supply_id, mpan_cores)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        f.write(msg)
    finally:
        if f is not None:
            f.close()


def _process_era(caches, sess, era_id, date, year_start):
    NORMAL_READ_TYPES = ("N", "C", "N3")

    era, supply, generator_type = (
        sess.query(Era, Supply, GeneratorType)
        .join(Supply, Era.supply_id == Supply.id)
        .outerjoin(GeneratorType, Supply.generator_type_id == GeneratorType.id)
        .filter(Era.id == era_id)
        .options(
            joinedload(Era.channels),
            joinedload(Era.cop),
            joinedload(Era.comm),
            joinedload(Era.dc_contract),
            joinedload(Era.exp_llfc),
            joinedload(Era.exp_supplier_contract),
            joinedload(Era.imp_llfc),
            joinedload(Era.imp_supplier_contract),
            joinedload(Era.mop_contract),
            joinedload(Era.mtc_participant).joinedload(MtcParticipant.mtc),
            joinedload(Era.mtc_participant).joinedload(MtcParticipant.meter_type),
            joinedload(Era.pc),
            joinedload(Era.site_eras).joinedload(SiteEra.site),
            joinedload(Era.ssc),
            joinedload(Era.energisation_status),
            joinedload(Era.supply).joinedload(Supply.source),
            joinedload(Era.supply).joinedload(Supply.gsp_group),
            joinedload(Era.supply).joinedload(Supply.dno),
        )
        .one()
    )

    site_codes = []
    site_names = []
    for site_era in era.site_eras:
        if site_era.is_physical:
            physical_site = site_era.site
        else:
            site = site_era.site
            site_codes.append(site.code)
            site_names.append(site.name)

    sup_eras = (
        sess.query(Era).filter(Era.supply == supply).order_by(Era.start_date).all()
    )
    supply_start_date = sup_eras[0].start_date
    supply_finish_date = sup_eras[-1].finish_date

    if era.imp_mpan_core is None:
        voltage_level_code = era.exp_llfc.voltage_level.code
        is_substation = era.exp_llfc.is_substation
    else:
        voltage_level_code = era.imp_llfc.voltage_level.code
        is_substation = era.imp_llfc.is_substation

    if generator_type is None:
        generator_type_str = ""
    else:
        generator_type_str = generator_type.code

    metering_type = era.meter_category

    if metering_type in ("nhh", "amr"):
        latest_prev_normal_read = (
            sess.query(RegisterRead)
            .join(Bill)
            .join(RegisterRead.previous_type)
            .filter(
                ReadType.code.in_(NORMAL_READ_TYPES),
                RegisterRead.previous_date <= date,
                Bill.supply_id == supply.id,
            )
            .order_by(RegisterRead.previous_date.desc())
            .options(joinedload(RegisterRead.previous_type))
            .first()
        )

        latest_pres_normal_read = (
            sess.query(RegisterRead)
            .join(Bill)
            .join(RegisterRead.present_type)
            .filter(
                ReadType.code.in_(NORMAL_READ_TYPES),
                RegisterRead.present_date <= date,
                Bill.supply == supply,
            )
            .order_by(RegisterRead.present_date.desc())
            .options(joinedload(RegisterRead.present_type))
            .first()
        )

        if latest_prev_normal_read is None and latest_pres_normal_read is None:
            latest_normal_read_date = None
            latest_normal_read_type = None
        elif latest_pres_normal_read is not None and latest_prev_normal_read is None:
            latest_normal_read_date = latest_pres_normal_read.present_date
            latest_normal_read_type = latest_pres_normal_read.present_type.code
        elif latest_pres_normal_read is None and latest_prev_normal_read is not None:
            latest_normal_read_date = latest_prev_normal_read.previous_date
            latest_normal_read_type = latest_prev_normal_read.previous_type.code
        elif (
            latest_pres_normal_read.present_date > latest_prev_normal_read.previous_date
        ):
            latest_normal_read_date = latest_pres_normal_read.present_date
            latest_normal_read_type = latest_pres_normal_read.present_type.code
        else:
            latest_normal_read_date = latest_prev_normal_read.previous_date
            latest_normal_read_type = latest_prev_normal_read.previous_type.code
        if latest_normal_read_date is not None:
            latest_normal_read_date = hh_format(latest_normal_read_date)

    else:
        latest_normal_read_date = metering_type
        latest_normal_read_type = None

    mop_contract = era.mop_contract
    mop_contract_name = mop_contract.name
    mop_account = era.mop_account
    latest_mop_bill_date = (
        sess.query(Bill.finish_date)
        .join(Batch)
        .filter(
            Bill.start_date <= date,
            Bill.supply == supply,
            Batch.contract == mop_contract,
        )
        .order_by(Bill.finish_date.desc())
        .first()
    )

    if latest_mop_bill_date is not None:
        latest_mop_bill_date = hh_format(latest_mop_bill_date[0])

    dc_contract = era.dc_contract
    dc_contract_name = dc_contract.name
    dc_account = era.dc_account
    latest_dc_bill_date = (
        sess.query(Bill.finish_date)
        .join(Batch)
        .filter(
            Bill.start_date <= date,
            Bill.supply == supply,
            Batch.contract == dc_contract,
        )
        .order_by(Bill.finish_date.desc())
        .first()
    )

    if latest_dc_bill_date is not None:
        latest_dc_bill_date = hh_format(latest_dc_bill_date[0])

    channel_values = []
    for imp_related in [True, False]:
        for channel_type in CHANNEL_TYPES:
            if era.find_channel(sess, imp_related, channel_type) is None:
                channel_values.append("false")
            else:
                channel_values.append("true")

    imp_avg_months = None
    exp_avg_months = None
    for is_import in [True, False]:
        if metering_type == "nhh":
            continue

        params = {
            "supply_id": supply.id,
            "year_start": year_start,
            "year_finish": date,
            "is_import": is_import,
        }
        month_mds = tuple(
            md[0] * 2
            for md in sess.execute(
                text(
                    """

select max(hh_datum.value) as md
from hh_datum join channel on (hh_datum.channel_id = channel.id)
join era on (channel.era_id = era.id)
where era.supply_id = :supply_id and hh_datum.start_date >= :year_start
and hh_datum.start_date <= :year_finish
and channel.channel_type = 'ACTIVE'
and channel.imp_related = :is_import
group by extract(month from (hh_datum.start_date at time zone 'utc'))
order by md desc
limit 3

"""
                ),
                params=params,
            )
        )

        avg_months = sum(month_mds)
        if len(month_mds) > 0:
            avg_months /= len(month_mds)
            if is_import:
                imp_avg_months = avg_months
            else:
                exp_avg_months = avg_months

    if (imp_avg_months is not None and imp_avg_months > 100) or (
        exp_avg_months is not None and exp_avg_months > 100
    ):
        mandatory_hh = "yes"
    else:
        mandatory_hh = "no"

    imp_latest_supplier_bill_date = None
    exp_latest_supplier_bill_date = None
    imp_tnuos_band = None
    for is_import in (True, False):
        for er in (
            sess.query(Era)
            .filter(Era.supply == era.supply, Era.start_date <= date)
            .order_by(Era.start_date.desc())
        ):
            if is_import:
                if er.imp_mpan_core is None:
                    break
                else:
                    supplier_contract = er.imp_supplier_contract
                    if era.energisation_status.code == "E":
                        imp_ss = SupplySource(
                            sess, date, date, date + HH, era, True, caches
                        )
                        chellow.e.duos.duos_vb(imp_ss)
                        hh = imp_ss.hh_data[0]
                        if hh["start-date"] >= chellow.e.tnuos.BANDED_START:
                            imp_tnuos_band = chellow.e.tnuos.BAND_LOOKUP[
                                hh["duos-description"]
                            ]
            else:
                if er.exp_mpan_core is None:
                    break
                else:
                    supplier_contract = er.exp_supplier_contract

            latest_bill_date = (
                sess.query(Bill.finish_date)
                .join(Batch)
                .filter(
                    Bill.finish_date >= er.start_date,
                    Bill.finish_date <= hh_min(er.finish_date, date),
                    Bill.supply == supply,
                    Batch.contract == supplier_contract,
                )
                .order_by(Bill.finish_date.desc())
                .first()
            )

            if latest_bill_date is not None:
                latest_bill_date = hh_format(latest_bill_date[0])

                if is_import:
                    imp_latest_supplier_bill_date = latest_bill_date
                else:
                    exp_latest_supplier_bill_date = latest_bill_date
                break

    meter_installation_date = (
        sess.query(func.min(Era.start_date))
        .filter(Era.supply == era.supply, Era.msn == era.msn)
        .one()[0]
    )

    ssc = era.ssc
    if ssc is None:
        ssc_code = ssc_description = num_registers = None
    else:
        ssc_code, ssc_description = ssc.code, ssc.description
        num_registers = (
            sess.query(MeasurementRequirement)
            .filter(MeasurementRequirement.ssc == ssc)
            .count()
        )

    return (
        [
            date,
            era.imp_mpan_core,
            era.exp_mpan_core,
            physical_site.code,
            physical_site.name,
            ", ".join(site_codes),
            ", ".join(site_names),
            supply.id,
            supply.source.code,
            generator_type_str,
            supply.gsp_group.code,
            supply.dno.dno_code,
            voltage_level_code,
            is_substation,
            metering_type,
            mandatory_hh,
            era.pc.code,
            era.mtc_participant.mtc.code,
            era.cop.code,
            era.comm.code,
            ssc_code,
            ssc_description,
            era.energisation_status.code,
            num_registers,
            mop_contract_name,
            mop_account,
            dc_contract_name,
            dc_account,
            era.msn,
            meter_installation_date,
            latest_normal_read_date,
            latest_normal_read_type,
            latest_dc_bill_date,
            latest_mop_bill_date,
            supply_start_date,
            supply_finish_date,
            None if era.dtc_meter_type is None else era.dtc_meter_type.code,
            imp_tnuos_band,
        ]
        + channel_values
        + [
            era.imp_sc,
            None if era.imp_llfc is None else era.imp_llfc.code,
            None if era.imp_llfc is None else era.imp_llfc.description,
            (
                None
                if era.imp_supplier_contract is None
                else era.imp_supplier_contract.name
            ),
            era.imp_supplier_account,
            imp_avg_months,
            imp_latest_supplier_bill_date,
        ]
        + [
            era.exp_sc,
            None if era.exp_llfc is None else era.exp_llfc.code,
            None if era.exp_llfc is None else era.exp_llfc.description,
            (
                None
                if era.exp_supplier_contract is None
                else era.exp_supplier_contract.name
            ),
            era.exp_supplier_account,
            exp_avg_months,
            exp_latest_supplier_bill_date,
        ]
    )


def _process(sess, f, date, supply_id, mpan_cores):
    writer = csv.writer(f, lineterminator="\n")
    titles = (
        "Date",
        "Import MPAN Core",
        "Export MPAN Core",
        "Physical Site Id",
        "Physical Site Name",
        "Other Site Ids",
        "Other Site Names",
        "Supply Id",
        "Source",
        "Generator Type",
        "GSP Group",
        "DNO Name",
        "Voltage Level",
        "Is Substations",
        "Metering Type",
        "Mandatory HH",
        "PC",
        "MTC",
        "CoP",
        "Comms Type",
        "SSC Code",
        "SSC Description",
        "Energisation Status",
        "Number Of Registers",
        "MOP Contract",
        "Mop Account",
        "DC Contract",
        "DC Account",
        "Meter Serial Number",
        "Meter Installation Date",
        "Latest Normal Meter Read Date",
        "Latest Normal Meter Read Type",
        "Latest DC Bill Date",
        "Latest MOP Bill Date",
        "Supply Start Date",
        "Supply Finish Date",
        "DTC Meter Type",
        "tnuos_band",
        "Import ACTIVE?",
        "Import REACTIVE_IMPORT?",
        "Import REACTIVE_EXPORT?",
        "Export ACTIVE?",
        "Export REACTIVE_IMPORT?",
        "Export REACTIVE_EXPORT?",
        "Import Agreed Supply Capacity (kVA)",
        "Import LLFC Code",
        "Import LLFC Description",
        "Import Supplier Contract",
        "Import Supplier Account",
        "Import Mandatory kW",
        "Latest Import Supplier Bill Date",
        "Export Agreed Supply Capacity (kVA)",
        "Export LLFC Code",
        "Export LLFC Description",
        "Export Supplier Contract",
        "Export Supplier Account",
        "Export Mandatory kW",
        "Latest Export Supplier Bill Date",
    )
    writer.writerow(titles)

    year_start = date + HH - relativedelta(years=1)
    caches = {}

    era_ids = (
        sess.query(Era.id)
        .filter(
            Era.start_date <= date,
            or_(Era.finish_date == null(), Era.finish_date >= date),
        )
        .order_by(Era.supply_id)
    )

    if supply_id is not None:
        supply = Supply.get_by_id(sess, supply_id)

        era_ids = era_ids.filter(Era.supply == supply)

    if mpan_cores is not None:
        era_ids = era_ids.filter(
            or_(Era.imp_mpan_core.in_(mpan_cores), Era.exp_mpan_core.in_(mpan_cores))
        )

    for (era_id,) in era_ids:
        try:
            vals = _process_era(caches, sess, era_id, date, year_start)
        except BaseException as e:
            raise BadRequest(f"Problem with era {era_id}: {e}") from e
        except BadRequest as e:
            raise BadRequest(f"Problem with era {era_id}: {e.description}")
        writer.writerow([csv_make_val(v) for v in vals])

        # Avoid a long-running transaction
        sess.rollback()


def do_get(session):
    user = g.user
    date = req_date("date")
    if "supply_id" in request.values:
        supply_id = req_int("supply_id")
    else:
        supply_id = None

    if "mpan_cores" in request.values:
        mpan_cores_str = req_str("mpan_cores")
        mpan_cores = mpan_cores_str.splitlines()
        if len(mpan_cores) == 0:
            mpan_cores = None
        else:
            for i in range(len(mpan_cores)):
                mpan_cores[i] = parse_mpan_core(mpan_cores[i])
    else:
        mpan_cores = None

    args = user.id, date, supply_id, mpan_cores
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal

from chellow.gas.dn_rate_parser import (
    State,
    _handle_EXIT_CAPACITY,
    _handle_SECTION,
    find_dn_rates,
    tab_gdn_unit_rates,
)


def test_handle_EXIT_CAPACITY_first_row(mocker):
    state = State.EXIT_CAPACITY
    b1 = mocker.Mock()
    b1.value = "EXIT CAPACITY UNIT RATES BY EXIT ZONE"
    sheet = {"B1": b1}
    row = 1
    rates = {}
    networks = {"EE": "C"}

    _handle_EXIT_CAPACITY(state, sheet, row, rates, networks)

    assert rates == {}


def test_handle_EXIT_CAPACITY_rate_row(mocker):
    state = State.EXIT_CAPACITY
    b1 = mocker.Mock()
    b1.value = "EA1"
    c1 = mocker.Mock()
    c1.value = 0.2
    sheet = {"B1": b1, "C1": c1}
    row = 1
    rates = {"exit_zones": {}}
    networks = {"EE": "C"}

    _handle_EXIT_CAPACITY(state, sheet, row, rates, networks)

    expected = {
        "exit_zones": {"EA1": {"exit_capacity_gbp_per_kwh_per_day": Decimal("0.002")}}
    }
    assert rates == expected


def test_handle_SECTION_system_capacity_to_73200(mocker):
    state = State.SYSTEM_CAPACITY
    b1 = mocker.Mock()
    b1.value = "UP TO 73,200 KWH PER ANNUM"
    c1 = mocker.Mock()
    c1.value = 0.2
    sheet = {"B1": b1, "C1": c1}
    row = 1
    rates = {"gdn": {"EE": {"system_capacity": {}}}}
    networks = {"EE": "C"}

    _handle_SECTION(state, sheet, row, rates, networks)

    expected = {
        "gdn": {
            "EE": {
                "system_capacity": {"to_73200_gbp_per_kwh_per_day": Decimal("0.002")}
            }
        }
    }
    assert rates == expected


def test_handle_SECTION_system_capacity_73200_and_over(mocker):
    state = State.SYSTEM_CAPACITY
    b1 = mocker.Mock()
    b1.value = "732,000 KWH PER ANNUM AND ABOVE"
    c1 = mocker.Mock()
    c1.value = 0.2
    c3 = mocker.Mock()
    c3.value = 0.6
    sheet = {"B1": b1, "C1": c1, "C3": c3}
    row = 1
    rates = {"gdn": {"EE": {"system_capacity": {"732000_and_over": {}}}}}
    networks = {"EE": "C"}

    _handle_SECTION(state, sheet, row, rates, networks)

    expected = {
        "gdn": {
            "EE": {
                "system_capacity": {
                    "732000_and_over": {
                        "gbp_per_kwh_per_day": Decimal("0.002000"),
                        "exponent": Decimal("0.6000"),
                    }
                }
            }
        }
    }
    assert rates == expected


def test_handle_SECTION_system_capacity_minimum(mocker):
    state = State.SYSTEM_CAPACITY
    b1 = mocker.Mock()
    b1.value = "SUBJECT TO A MINIMUM RATE OF"
    c1 = mocker.Mock()
    c1.value = 0.2
    sheet = {"B1": b1, "C1": c1}
    row = 1
    rates = {"gdn": {"EE": {"system_capacity": {"732000_and_over": {}}}}}
    networks = {"EE": "C"}

    _handle_SECTION(state, sheet, row, rates, networks)

    expected = {
        "gdn": {
            "EE": {
                "system_capacity": {
                    "732000_and_over": {
                        "minimum_gbp_per_kwh_per_day": Decimal("0.002"),
                    }
                }
            }
        }
    }
    assert rates == expected


def test_tab_gdn_unit_rates(mocker):
    b1 = mocker.Mock()
    b1.value = "NETWORK"
    b1.column_letter = "B"
    c1 = mocker.Mock()
    c1.value = "EAST OF ENGLAND"
    c1.column_letter = "C"
    b2 = mocker.Mock()
    b2.value = "EXIT CAPACITY UNIT RATES BY EXIT ZONE"
    b2.column_letter = "B"
    b3 = mocker.Mock()
    b3.value = None
    b3.column_letter = "B"
    sheet = {
        "B1": b1,
        "C1": c1,
        "B2": b2,
        "B3": b3,
        "B": [b1, b2, b3],
        1: [None, b1, c1],
    }
    rates = {"gdn": {}}
    tab_gdn_unit_rates(sheet, rates)


def test_find_dn_rates(mocker):
    mock_book = mocker.Mock()
    mock_book.worksheets = []
    mocker.patch(
        "chellow.gas.dn_rate_parser.openpyxl.load_workbook", return_value=mock_book
    )
    file_name = "fname.xls"
    file_like = mocker.Mock()
    rates = find_dn_rates(file_name, file_like)
    assert rates == {"a_file_name": file_name, "exit_zones": {}, "gdn": {}}



import csv
import threading
import traceback

from flask import g

from sqlalchemy.sql import func, select

from chellow.dloads import open_file
from chellow.models import (
    Batch,
    Bill,
    GBatch,
    GBill,
    Session,
    User,
)
from chellow.utils import csv_make_val
from chellow.views import chellow_redirect


def content(user_id):
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("batches.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")

            titles = (
                "utility",
                "chellow_id",
                "reference",
                "description",
                "contract_name",
                "num_bills",
                "net_gbp",
                "vat_gbp",
                "gross_gbp",
                "kwh",
                "min_start_date",
                "max_finish_date",
            )
            writer.writerow(titles)

            for batch in sess.execute(
                select(Batch).order_by(Batch.contract_id, Batch.reference)
            ).scalars():
                (
                    num_bills,
                    sum_net_gbp,
                    sum_vat_gbp,
                    sum_gross_gbp,
                    sum_kwh,
                    min_start_date,
                    max_finish_date,
                ) = sess.execute(
                    select(
                        func.count(Bill.id),
                        func.sum(Bill.net),
                        func.sum(Bill.vat),
                        func.sum(Bill.gross),
                        func.sum(Bill.kwh),
                        func.min(Bill.start_date),
                        func.max(Bill.finish_date),
                    ).where(Bill.batch == batch)
                ).one()

                if sum_net_gbp is None:
                    sum_net_gbp = sum_vat_gbp = sum_gross_gbp = sum_kwh = 0
                vals = {
                    "utility": "electricity",
                    "chellow_id": batch.id,
                    "reference": batch.reference,
                    "description": batch.description,
                    "contract_name": batch.contract.name,
                    "num_bills": num_bills,
                    "net_gbp": sum_net_gbp,
                    "vat_gbp": sum_vat_gbp,
                    "gross_gbp": sum_gross_gbp,
                    "kwh": sum_kwh,
                    "min_start_date": min_start_date,
                    "max_finish_date": max_finish_date,
                }

                writer.writerow(csv_make_val(vals[t]) for t in titles)

                # Avoid a long-running transaction
                sess.rollback()

            for g_batch in sess.execute(
                select(GBatch).order_by(GBatch.g_contract_id, GBatch.reference)
            ).scalars():
                (
                    num_bills,
                    sum_net_gbp,
                    sum_vat_gbp,
                    sum_gross_gbp,
                    sum_kwh,
                    min_start_date,
                    max_finish_date,
                ) = sess.execute(
                    select(
                        func.count(GBill.id),
                        func.sum(GBill.net),
                        func.sum(GBill.vat),
                        func.sum(GBill.gross),
                        func.sum(GBill.kwh),
                        func.min(GBill.start_date),
                        func.max(GBill.finish_date),
                    ).where(GBill.g_batch == g_batch)
                ).one()

                if sum_net_gbp is None:
                    sum_net_gbp = sum_vat_gbp = sum_gross_gbp = sum_kwh = 0
                vals = {
                    "utility": "gas",
                    "chellow_id": g_batch.id,
                    "reference": g_batch.reference,
                    "description": g_batch.description,
                    "contract_name": g_batch.g_contract.name,
                    "num_bills": num_bills,
                    "net_gbp": sum_net_gbp,
                    "vat_gbp": sum_vat_gbp,
                    "gross_gbp": sum_gross_gbp,
                    "kwh": sum_kwh,
                    "min_start_date": min_start_date,
                    "max_finish_date": max_finish_date,
                }

                writer.writerow(csv_make_val(vals[t]) for t in titles)

                # Avoid a long-running transaction
                sess.rollback()

    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    threading.Thread(target=content, args=(g.user.id,)).start()
    return chellow_redirect("/downloads", 303)
